from lto.accounts.account_factory import AccountFactory
from ecdsa import VerifyingKey, SECP256k1, NIST256p, SigningKey
import hashlib
from ecdsa.util import randrange_from_seed__trytryagain as randrange_from_seed
import base58
from lto import crypto
from lto.accounts.ecdsa.account_ecdsa import AccountECDSA as Account
from lto.accounts.brainwallet import random_seed as brainwallet_random_seed

from mnemonic import Mnemonic
from lto import ethereum_mnemonic_utils as eth


class AccountFactoryECDSA(AccountFactory):

    def __init__(self, chain_id, curve='secp256k1'):
        super().__init__(chain_id, 'bip39')

        self.key_type = curve
        if curve == 'secp256k1':
            self.curve = SECP256k1
        elif curve == 'secp256r1':
            self.curve = NIST256p
        else:
            raise Exception("Curve not supported")

    def _MakeKey(self, seed):
        secexp = randrange_from_seed(seed, self.curve.order)
        return SigningKey.from_secret_exponent(secexp, curve=self.curve, hashfunc=hashlib.sha256)

    def create_sign_keys(self, seed, nonce=0):
        private_key = self._MakeKey(seed)
        public_key = private_key.verifying_key
        return private_key, public_key, self.key_type


    def create_address(self, public_key):
        unhashed_address = chr(1) + str(self.chain_id) + crypto.hash_chain(public_key.to_string(encoding="compressed"))[0:20]
        address_hash = crypto.hash_chain(crypto.str2bytes(unhashed_address))[0:4]
        return base58.b58encode(crypto.str2bytes(unhashed_address + address_hash))

    def create_from_public_key(self, public_key):
        if not isinstance(public_key, VerifyingKey):
            if isinstance(public_key, bytes):
                public_key = VerifyingKey.from_string(public_key, curve=self.curve, hashfunc=hashlib.sha256)
            elif isinstance(public_key, str):
                public_key = base58.b58decode(public_key)
                public_key = VerifyingKey.from_string(public_key, curve=self.curve, hashfunc=hashlib.sha256)
            else:
                raise Exception("Unrecognized Public Key format")
        address = self.create_address(public_key)
        return Account(address=address, public_key=public_key, key_type=self.key_type)

    def create_from_private_key(self, private_key):
        if not isinstance(private_key, SigningKey):
            if isinstance(private_key, bytes):
                private_key = SigningKey.from_string(private_key, curve=self.curve, hashfunc=hashlib.sha256)
            elif isinstance(private_key, str):
                private_key = base58.b58decode(private_key)
                private_key = SigningKey.from_string(private_key, curve=self.curve, hashfunc=hashlib.sha256)
            else:
                raise Exception("Unrecognized Private Key format")
        public_key = private_key.verifying_key
        address = self.create_address(public_key)
        return Account(address=address, public_key=public_key, private_key=private_key, key_type=self.key_type)

    def create_from_seed(self, seed, nonce=0):
        raise Exception ('Method under construction')
        '''
        private_key = eth.mnemonic_to_private_key(seed, nonce=nonce)
        public_key = eth.derive_public_key(private_key)
        address = eth.address_from_private_key(private_key)
        key_type = "secp256k1"
        # private_key, public_key, key_type = self.create_sign_keys(seed, nonce)
        # address = self.create_address(public_key)
        return Account(address[2:], public_key, private_key, key_type, seed, nonce)'''

    def create_with_values(self, address, public_key, private_key, key_type, seed=None):
        return Account(address, public_key, private_key, key_type, seed)

    def create(self):
        seed = brainwallet_random_seed()
        private_key, public_key, key_type = self.create_sign_keys(seed, 0)
        address = self.create_address(public_key)
        return Account(address, public_key, private_key, key_type, seed, 0)



import requests
import json

from lto.transactions import from_data as tx_from_data, SetScript
from lto.accounts import Account
from lto import crypto


class PublicNode(object):
    def __init__(self, url, api_key=''):
        self.url = url
        self.api_key = api_key
        
    @staticmethod
    def __addr(account_or_address):
        return account_or_address.address if isinstance(account_or_address, Account) else account_or_address

    def wrapper(self, api, post_data='', host='', headers=None):
        if headers is None:
            headers = {}

        if not host:
            host = self.url

        if self.api_key:
            headers = {"X-API-Key": self.api_key}

        if post_data:
            r = requests.post('%s%s' % (host, api), data=post_data,
                              headers=crypto.merge_dicts(headers, {'content-type': 'application/json'}))
        else:
            r = requests.get('%s%s' % (host, api), headers=headers)

        if r.status_code != 200:
            method = 'POST' if post_data else 'GET'
            try:
                error = json.loads(r.text)
            except:
                error = r.text
            
            raise Exception(
                '{} {}{} responded with {} {}'.format(method, host, api, r.status_code, r.reason),
                error
            )

        r.raise_for_status()

        return r.json()

    def broadcast(self, transaction):
        data = json.dumps(transaction.to_json())
        response = self.wrapper(api='/transactions/broadcast', post_data=data)
        return tx_from_data(response)

    def compile(self, script_source):
        compiled_script = self.wrapper(api='/utils/script/compile', post_data=script_source)['script']
        return SetScript(compiled_script)

    def height(self):
        return self.wrapper('/blocks/height')['height']

    def last_block(self):
        return self.wrapper('/blocks/last')

    def block(self, n):
        return self.wrapper('/blocks/at/%d' % n)

    def tx(self, id):
        response = self.wrapper('/transactions/info/%s' % id)
        return tx_from_data(response)

    def lease_list(self, address):
        return self.wrapper(api='/leasing/active/{}'.format(self.__addr(address)))

    def get_data(self, address):
        return self.wrapper(api='/addresses/data/{}'.format(self.__addr(address)))

    def get_data_by_key(self, address, key):
        return self.wrapper(api='/addresses/data/{}/{}'.format(self.__addr(address), key))

    def sponsorship_list(self, address):
        return self.wrapper(api='/sponsorship/status/{}'.format(self.__addr(address)))

    def association_list(self, address):
        return self.wrapper(api='/associations/status/{}'.format(self.__addr(address)))

    def node_status(self):
        return self.wrapper(api='/node/status')

    def balance(self, address):
        try:
            return self.wrapper('/addresses/balance/%s' % self.__addr(address))['balance']
        except:
            return -1

    def balance_details(self, address):
        return self.wrapper('/addresses/balance/details/%s' % self.__addr(address))

    def validate_address(self, address):
        return self.wrapper('/addresses/validate/{}'.format(address))['valid']


    def data_of(self, address):
        data = self.wrapper('/addresses/data/%s' % self.__addr(address))
        dict = {}
        for entry in data:
            dict[entry['key']] = entry['value']
        return dict

    def transactions(self, address, limit=100, after=''):
        return self.wrapper('/transactions/address/%s/limit/%d%s' % (
            self.__addr(address), limit, "" if after == "" else "?after={}".format(after)))

    def sign_transaction(self, transaction):
        data = json.dumps(transaction.to_json())
        return(self.wrapper(api='/transactions/sign', post_data=data))




import base58
from lto import crypto
from lto.transaction import Transaction
import struct


class Anchor(Transaction):
    TYPE = 15
    BASE_FEE = 25000000
    VAR_FEE = 10000000
    DEFAULT_VERSION = 3

    def __init__(self, *anchors):
        super().__init__()

        self.anchors = anchors
        self.tx_fee = self.BASE_FEE + len(anchors) * self.VAR_FEE
        self.version = self.DEFAULT_VERSION

    def __anchors_to_binary(self):
        binary = b''
        for anchor in self.anchors:
            binary += struct.pack(">H", len(crypto.str2bytes(anchor)))
            binary += crypto.str2bytes(anchor)

        return binary

    def __to_binary_v1(self):
        return (self.TYPE.to_bytes(1, 'big') +
                b'\1' +
                base58.b58decode(self.sender_public_key) +
                struct.pack(">H", len(self.anchors)) +
                self.__anchors_to_binary() +
                struct.pack(">Q", self.timestamp) +
                struct.pack(">Q", self.tx_fee))

    def __to_binary_v3(self):
        return (self.TYPE.to_bytes(1, 'big') +
                b'\3' +
                crypto.str2bytes(self.chain_id) +
                struct.pack(">Q", self.timestamp) +
                crypto.key_type_id(self.sender_key_type) +
                base58.b58decode(self.sender_public_key) +
                struct.pack(">Q", self.tx_fee) +
                struct.pack(">H", len(self.anchors)) +
                self.__anchors_to_binary())

    def to_binary(self):
        if self.version == 1:
            return self.__to_binary_v1()
        elif self.version == 3:
            return self.__to_binary_v3()
        else:
            raise Exception('Incorrect Version')

    def to_json(self):
        return (crypto.merge_dicts(
            {
                "id": self.id if self.id else "",
                "type": self.TYPE,
                "version": self.version,
                "sender": self.sender,
                "senderKeyType": self.sender_key_type,
                "senderPublicKey": self.sender_public_key,
                "fee": self.tx_fee,
                "timestamp": self.timestamp,
                "anchors": list(map(lambda anchor: base58.b58encode(crypto.str2bytes(anchor)), self.anchors)),
                "proofs": self.proofs,
                "height": self.height if self.height else ""
            },
            self._sponsor_json()))

    @staticmethod
    def from_data(data):
        tx = Anchor("")
        tx.id = data['id'] if 'id' in data else ''
        tx.type = data['type']
        tx.version = data['version']
        tx.sender = data['sender'] if 'sender' in data else ''
        tx.sender_key_type = data['senderKeyType'] if 'senderKeyType' in data else 'ed25519'
        tx.sender_public_key = data['senderPublicKey']
        tx.fee = data['fee']
        tx.timestamp = data['timestamp']
        tx.anchors = data['anchors']
        tx.proofs = data['proofs'] if 'proofs' in data else []
        tx.height = data['height'] if 'height' in data else ''

        if "sponsor_public_key" in data:
            tx.sponsor = data['sponsor']
            tx.sponsor_public_key = data['sponsorPublicKey']
            tx.sponsor_key_type = data['sponsorKeyType']

        return tx



from abc import ABC, abstractmethod
from time import time
import base58
from nacl.signing import VerifyKey


class Transaction(ABC):

    def __init__(self):
        self.tx_fee = 0
        self.timestamp = 0
        self.proofs = []
        self.sender = ''
        self.sender_public_key = ''
        self.chain_id = ''
        self.sponsor = ''
        self.sponsor_public_key = ''
        self.sender_key_type = 'ed25519'
        self.sponsor_key_type = 'ed25519'
        self.id = None
        self.height = None

    @abstractmethod
    def to_binary(self):
        pass

    def is_signed(self):
        return len(self.proofs) != 0

    def sign_with(self, account):
        if self.timestamp == 0:
            self.timestamp = int(time() * 1000)

        if self.sender == '':
            self.sender = account.address
            self.sender_public_key = account.get_public_key()

        self.chain_id = account.get_network()
        self.sender_key_type = account.key_type

        self.proofs.append(account.sign(self.to_binary()))

    def sponsor_with(self, sponsor_account):
        if not self.is_signed():
            raise Exception('Transaction must be signed first')

        self.sponsor = sponsor_account.address
        self.sponsor_public_key = sponsor_account.get_public_key()
        self.sponsor_key_type = sponsor_account.key_type
        self.proofs.append(sponsor_account.sign(self.to_binary()))

    def broadcast_to(self, node):
        return node.broadcast(self)

    @abstractmethod
    def to_json(self):
        pass

    def _sponsor_json(self):
        if self.sponsor:
            return {"sponsor": self.sponsor,
                    "sponsorPublicKey": self.sponsor_public_key,
                    "sponsorKeyType": self.sponsor_key_type}
        else:
            return {}




from lto import crypto
from abc import ABC, abstractmethod


class Account(ABC):

    def __init__(self, address, public_key, private_key=None, key_type='ed25519', seed=None, nonce=0):
        self.address = address
        self.public_key = public_key
        self.private_key = private_key
        self.seed = seed
        self.nonce = nonce
        self.key_type = key_type

    @abstractmethod
    def get_public_key(self):
        pass

    @abstractmethod
    def sign(self, message):
        pass

    @abstractmethod
    def verify_signature(self, message: str, signature: str, encoding: str = 'base58'):
        pass

    def get_network(self):
        return crypto.get_network(self.address)



import copy

from lto.accounts.account_factory import AccountFactory
from lto import crypto
import struct
from nacl.signing import SigningKey, VerifyKey
import base58
from lto.accounts.account import Account
from lto.accounts.ed25519.account_ed25519 import AccountED25519 as Account
from mnemonic import Mnemonic


class AccountFactoryED25519(AccountFactory):

    def __init__(self, chain_id, seed_method='brainwallet'):
        super().__init__(chain_id, seed_method)
        self.key_type = 'ed25519'
        self.mnemo = self.__create_mnemo(seed_method)

    def with_seed_method(self, seed_method):
        clone = super().with_seed_method(seed_method)
        if clone != self:
            self.mnemo = self.__create_mnemo(seed_method)

        return clone

    def create_sign_keys(self, seed, nonce=0):
        if self.seed_method == 'brainwallet':
            seed_hash = crypto.hash_chain(struct.pack(">L", nonce) + crypto.str2bytes(seed))
            account_seed_hash = crypto.sha256(seed_hash)
        elif self.seed_method == 'bip39' or self.seed_method.startswith('bip39:'):
            raise Exception('Seed method under construction')
            #account_seed_hash = self.mnemo.to_seed(seed)
        else:
            raise Exception('Unsupported seed method')
        private_key = SigningKey(account_seed_hash[:32])
        public_key = private_key.verify_key
        return private_key, public_key, self.key_type

    def create_address(self, public_key):
        unhashed_address = chr(1) + str(self.chain_id) + crypto.hash_chain(public_key.__bytes__())[0:20]
        address_hash = crypto.hash_chain(crypto.str2bytes(unhashed_address))[0:4]
        return base58.b58encode(crypto.str2bytes(unhashed_address + address_hash))

    def create_from_private_key(self, private_key):
        if not isinstance(private_key, SigningKey):
            public_key = VerifyKey(base58.b58decode(private_key)[-32:])
            private_key = SigningKey(seed=base58.b58decode(private_key)[:-32])
            address = self.create_address(public_key)
        else:
            public_key = private_key.verify_key
            address = self.create_address(public_key)
        return Account(address, public_key, private_key, self.key_type)

    def create_from_public_key(self, public_key):
        if isinstance(public_key, bytes):
            public_key = VerifyKey(public_key)
        elif isinstance(public_key, str):
            public_key = VerifyKey(base58.b58decode(public_key))

        if not isinstance(public_key, VerifyKey):
            raise Exception("Unrecognized Public Key format")

        address = self.create_address(public_key)

        return Account(address=address, public_key=public_key, key_type=self.key_type)

    def create_from_seed(self, seed, nonce=0):
        private_key, public_key, key_type = self.create_sign_keys(seed, nonce)
        address = self.create_address(public_key)
        return Account(address, public_key, private_key, key_type, seed, nonce)

    def create_with_values(self, address, public_key, private_key, key_type, seed=None):
        return Account(address, public_key, private_key, key_type, seed)

    def __create_mnemo(self, seed_method):
        if seed_method == 'bip39':
            return Mnemonic("english")
        elif seed_method.startswith('bip39:'):
            return Mnemonic(seed_method[6:])
        else:
            return None


